using System.Diagnostics;
using System;
using System.Windows.Forms;
using System.Collections;
using System.Drawing;
using Microsoft.VisualBasic;
using System.Data;
using System.Collections.Generic;
using System.IO;
using Microsoft.VisualBasic.CompilerServices;


namespace Best_XY_MLR_cs
{
    public class CStatSum
    {

        private const double EPSILON = 1.0e-50;
        private const string DIGIT_MARKERS = "-+0123456789.";

        private bool m_bZeroX;
        private bool m_bZeroY;
        private bool m_bNegX;
        private bool m_bNegY;

        private double m_fSum;
        private double m_fSumX11;
        private double m_fSumX21;
        private double m_fSumX22;
        private double m_fSumX12;
        private double m_fSumY;
        private double m_fSumY2;
        private double m_fSumX1Y;
        private double m_fSumX2Y;
        private double m_fSumX1X2;
        private double m_fMeanX1;
        private double m_fMeanX2;
        private double m_fMeanY;
        private double m_fSdevX1;
        private double m_fSdevX2;
        private double m_fSdevY;
        private double m_fSlope1;
        private double m_fSlope2;
        private double m_fIntercept;
        private double m_fR2;
        private string m_sTX1;
        private string m_sTX2;
        private string m_sTY;
        private string m_sWt;

        private int m_nDataCount;
        private double[] m_fX;
        private double[] m_fY;
        private double[] m_fWt;


        public void InitSums()
        {
            m_fSum = 0;
            m_fSumX11 = 0;
            m_fSumX12 = 0;
            m_fSumX21 = 0;
            m_fSumX22 = 0;
            m_fSumY = 0;
            m_fSumY2 = 0;
            m_fSumX1Y = 0;
            m_fSumX2Y = 0;
            m_fSumX1X2 = 0;
            m_sTX1 = "";
            m_sTX2 = "";
            m_sTY = "";
            m_sWt = "";
        }

        public CStatSum()
        {
            InitSums();
        }

        public bool GetData(string sDataFilename, ref int nDataCOunt, double ShiftX, double ShiftY, double ScaleX, double ScaleY)
        {

            string sLine;
            string[] sLines;
            string[] sData;
            int I, J, K, N;
            bool bRes = true;
            char[] cDelim = new char[1];


            try
            {
                sLines = File.ReadAllLines(sDataFilename);

                nDataCOunt = sLines.GetUpperBound(0);

                // Dimension arrays for maximum capacity
                m_fX = new double[nDataCOunt + 1];
                m_fY = new double[nDataCOunt + 1];
                m_fWt = new double[nDataCOunt + 1];

                J = 0;
                m_nDataCount = 0;
                cDelim[0] = ',';
                while (!(J == sLines.Length))
                {
                    sLine = sLines[J].Trim(null);
                    // is line not empty?
                    if (sLine.Length > 0)
                    {
                        // is it NOT a comment?
                        if (DIGIT_MARKERS.IndexOf(sLine.Substring(0, 1)) >= 0)
                        {
                            sData = sLine.Split(cDelim);
                            N = sData.GetUpperBound(0);
                            m_fX[m_nDataCount] = double.Parse(sData[0]);
                            m_fY[m_nDataCount] = double.Parse(sData[1]);
                            if (N < 2)
                            {
                                m_fWt[m_nDataCount] = 1;
                            }
                            else
                            {
                                m_fWt[m_nDataCount] = double.Parse(sData[2]);
                            }
                            m_nDataCount++;
                        }
                    }
                    J++;
                }

                // adjust arrays to actual number of data
                m_fX = (double[])Microsoft.VisualBasic.CompilerServices.Utils.CopyArray((Array)m_fX, new double[m_nDataCount + 1]);
                m_fY = (double[])Microsoft.VisualBasic.CompilerServices.Utils.CopyArray((Array)m_fY, new double[m_nDataCount + 1]);
                m_fWt = (double[])Microsoft.VisualBasic.CompilerServices.Utils.CopyArray((Array)m_fWt, new double[m_nDataCount + 1]);
                nDataCOunt = m_nDataCount;

                for (I = 0; I <= m_nDataCount - 1; I++)
                {
                    m_fX[I] = ScaleX * (m_fX[I] - ShiftX);
                    m_fY[I] = ScaleY * (m_fY[I] - ShiftY);
                }

            }
            catch (Exception)
            {
                bRes = false;
            }

            return bRes;

        }

        private void Add(double X1, double X2, double Y, double Wt)
        {
            m_fSum = m_fSum + Wt;
            m_fSumX11 = m_fSumX11 + X1 * Wt;
            m_fSumX21 = m_fSumX21 + X2 * Wt;
            m_fSumX12 = m_fSumX12 + X1 * X1 * Wt;
            m_fSumX22 = m_fSumX22 + X2 * X2 * Wt;
            m_fSumY = m_fSumY + Y * Wt;
            m_fSumY2 = m_fSumY2 + Y * Y * Wt;
            m_fSumX1Y = m_fSumX1Y + X1 * Y * Wt;
            m_fSumX2Y = m_fSumX2Y + X2 * Y * Wt;
            m_fSumX1X2 = m_fSumX1X2 + X1 * X2 * Wt;
        }

        public void FindBestFit(ref CResults objRes)
        {
            int I;
            TypeModule.FitType ITX1;
            TypeModule.FitType ITX2;
            TypeModule.FitType ITY;
            bool bOK;
            double fXt1;
            double fXt2;
            double fYt;
            CErrors objErrs;
            string sModel;
            string sErr;

            objErrs = new CErrors();
            try
            {
                m_bZeroX = false;
                m_bZeroY = false;
                m_bNegX = false;
                m_bNegY = false;

                objRes.Clear();

                for (I = 0; I <= m_nDataCount - 1; I++)
                {
                    if (m_fX[I] < 0)
                    {
                        m_bNegX = true;
                    }
                    if (m_fY[I] < 0)
                    {
                        m_bNegY = true;
                    }
                    if (Math.Abs(m_fX[I]) < EPSILON)
                    {
                        m_bZeroX = true;
                    }
                    if (Math.Abs(m_fY[I]) < EPSILON)
                    {
                        m_bZeroY = true;
                    }
                }

                for (ITY = TypeModule.FitType.eLinear; ITY <= TypeModule.FitType.eLn; ITY++)
                {
                    // validate transformations
                    if (m_bZeroY && m_bNegY)
                    {
                        bOK = CanHandleZeroAndNegative(ITY);
                    }
                    else if (m_bZeroY)
                    {
                        bOK = CanHandleZero(ITY);
                    }
                    else if (m_bNegY)
                    {
                        bOK = CanHandleNegative(ITY);
                    }
                    else
                    {
                        bOK = true;
                    }

                    // Can proceed?
                    if (bOK)
                    {

                        for (ITX1 = TypeModule.FitType.eLinear; ITX1 <= TypeModule.FitType.eLn; ITX1++)
                        {

                            // validate transformations
                            if (m_bZeroX && m_bNegX)
                            {
                                bOK = CanHandleZeroAndNegative(ITX1);
                            }
                            else if (m_bZeroX)
                            {
                                bOK = CanHandleZero(ITX1);
                            }
                            else if (m_bNegX)
                            {
                                bOK = CanHandleNegative(ITX1);
                            }
                            else
                            {
                                bOK = true;
                            }

                            // Can proceed?
                            if (bOK)
                            {

                                for (ITX2 = ITX1; ITX2 <= TypeModule.FitType.eLn; ITX2++)
                                {

                                    // validate transformations
                                    if (m_bZeroX && m_bNegX)
                                    {
                                        bOK = CanHandleZeroAndNegative(ITX2);
                                    }
                                    else if (m_bZeroX)
                                    {
                                        bOK = CanHandleZero(ITX2);
                                    }
                                    else if (m_bNegX)
                                    {
                                        bOK = CanHandleNegative(ITX2);
                                    }
                                    else
                                    {
                                        bOK = true;
                                    }

                                    // lastly check if two transformations are the same
                                    if (ITX1 == ITX2)
                                    {
                                        bOK = false;
                                    }

                                    if (bOK)
                                    {

                                        // initialize summations
                                        InitSums();

                                        for (I = 0; I <= m_nDataCount - 1; I++)
                                        {
                                            fXt1 = 0;
                                            fXt2 = 0;
                                            fYt = 0;

                                            switch (ITX1)
                                            {
                                                case TypeModule.FitType.eLinear:

                                                    fXt1 = m_fX[I];
                                                    break;
                                                case TypeModule.FitType.eSquare:

                                                    fXt1 = Math.Pow(m_fX[I], 2);
                                                    break;
                                                case TypeModule.FitType.eCube:

                                                    fXt1 = Math.Pow(m_fX[I], 3);
                                                    break;
                                                case TypeModule.FitType.eCubeRoot:

                                                    fXt1 = Math.Pow(m_fX[I], (1 / 3));
                                                    break;
                                                case TypeModule.FitType.eRecip:

                                                    fXt1 = 1 / m_fX[I];
                                                    break;
                                                case TypeModule.FitType.eRecipCubeRoot:

                                                    fXt1 = Math.Pow(1 / m_fX[I], (1 / 3));
                                                    break;
                                                case TypeModule.FitType.eRecipSquare:

                                                    fXt1 = Math.Pow(1 / m_fX[I], 2);
                                                    break;
                                                case TypeModule.FitType.eRecipCube:

                                                    fXt1 = Math.Pow(1 / m_fX[I], 3);
                                                    break;
                                                case TypeModule.FitType.eSqrt:

                                                    fXt1 = Math.Sqrt(m_fX[I]);
                                                    break;
                                                case TypeModule.FitType.eRecipSqrt:

                                                    fXt1 = 1 / Math.Sqrt(m_fX[I]);
                                                    break;
                                                case TypeModule.FitType.eLn:

                                                    fXt1 = Math.Log(m_fX[I]);
                                                    break;
                                            }

                                            switch (ITX2)
                                            {
                                                case TypeModule.FitType.eLinear:

                                                    fXt2 = m_fX[I];
                                                    break;
                                                case TypeModule.FitType.eSquare:

                                                    fXt2 = Math.Pow(m_fX[I], 2);
                                                    break;
                                                case TypeModule.FitType.eCube:

                                                    fXt2 = Math.Pow(m_fX[I], 3);
                                                    break;
                                                case TypeModule.FitType.eCubeRoot:

                                                    fXt2 = Math.Pow(m_fX[I], (1 / 3));
                                                    break;
                                                case TypeModule.FitType.eRecip:

                                                    fXt2 = 1 / m_fX[I];
                                                    break;
                                                case TypeModule.FitType.eRecipCubeRoot:

                                                    fXt2 = Math.Pow(1 / m_fX[I], (1 / 3));
                                                    break;
                                                case TypeModule.FitType.eRecipSquare:

                                                    fXt2 = Math.Pow(1 / m_fX[I], 2);
                                                    break;
                                                case TypeModule.FitType.eRecipCube:

                                                    fXt2 = Math.Pow(1 / m_fX[I], 3);
                                                    break;
                                                case TypeModule.FitType.eSqrt:

                                                    fXt2 = Math.Sqrt(m_fX[I]);
                                                    break;
                                                case TypeModule.FitType.eRecipSqrt:

                                                    fXt2 = 1 / Math.Sqrt(m_fX[I]);
                                                    break;
                                                case TypeModule.FitType.eLn:

                                                    fXt2 = Math.Log(m_fX[I]);
                                                    break;
                                            }


                                            switch (ITY)
                                            {
                                                case TypeModule.FitType.eLinear:

                                                    fYt = m_fY[I];
                                                    break;
                                                case TypeModule.FitType.eSquare:

                                                    fYt = Math.Pow(m_fY[I], 2);
                                                    break;
                                                case TypeModule.FitType.eCube:

                                                    fYt = Math.Pow(m_fY[I], 3);
                                                    break;
                                                case TypeModule.FitType.eCubeRoot:

                                                    fYt = Math.Pow(m_fY[I], (1 / 3));
                                                    break;
                                                case TypeModule.FitType.eRecip:

                                                    fYt = 1 / m_fY[I];
                                                    break;
                                                case TypeModule.FitType.eRecipCubeRoot:

                                                    fYt = Math.Pow(1 / m_fY[I], (1 / 3));
                                                    break;
                                                case TypeModule.FitType.eRecipSquare:

                                                    fYt = Math.Pow(1 / m_fY[I], 2);
                                                    break;
                                                case TypeModule.FitType.eRecipCube:

                                                    fYt = Math.Pow(1 / m_fY[I], 3);
                                                    break;
                                                case TypeModule.FitType.eSqrt:

                                                    fYt = Math.Sqrt(m_fY[I]);
                                                    break;
                                                case TypeModule.FitType.eRecipSqrt:

                                                    fYt = 1 / Math.Sqrt(m_fY[I]);
                                                    break;
                                                case TypeModule.FitType.eLn:

                                                    fYt = Math.Log(m_fY[I]);
                                                    break;
                                            }

                                            // add transformed data to statistical summations
                                            Add(fXt1, fXt2, fYt, m_fWt[I]);

                                        }

                                        // store transformation data
                                        m_sTX1 = SayTransform(ITX1, "X");
                                        m_sTX2 = SayTransform(ITX2, "X");
                                        m_sTY = SayTransform(ITY, "Y");
                                        sModel = m_sTY + " - A + B1 * " + m_sTX1 + " + B2 * " + m_sTX2;
                                        // calculate regression statistics and store in
                                        // object accessed by m_objRes
                                        CalcLR(ref objRes, ref objErrs);

                                        if (objErrs.GetCount() > 0)
                                        {
                                            sErr = objErrs.GetErrText(0);
                                        }
                                        else
                                        {
                                            sErr = "";
                                        }
                                        objErrs.Clear(); // reset error object

                                        objRes.Add(sModel, m_fR2, m_fSlope1, m_fSlope2, m_fIntercept, sErr);

                                    }
                                    else
                                    {
                                        m_sTX1 = SayTransform(ITX1, "X");
                                        m_sTX2 = SayTransform(ITX2, "X");
                                        m_sTY = SayTransform(ITY, "Y");
                                        sModel = m_sTY + " - A + B1 * " + m_sTX1 + " + B2 * " + m_sTX2;
                                        // objRes.Add(sModel, -1, 0, 0, 0, 0)

                                    }

                                }

                            }

                        }

                    }
                }


            }
            catch (Exception ex)
            {
                objErrs.Add(ex.Message);
            }

        }

        private string SayTransform(TypeModule.FitType eVal, string sVar)
        {
            switch (eVal)
            {
                case TypeModule.FitType.eLinear:

                    return sVar;
                    break;
                case TypeModule.FitType.eSquare:
                    return sVar + "^2";
                    break;

                case TypeModule.FitType.eCube:
                    return sVar + "^3";
                    break;
                case TypeModule.FitType.eCubeRoot:
                    return sVar + "^1/3";
                    break;
                case TypeModule.FitType.eRecip:
                    return "1/" + sVar;
                    break;
                case TypeModule.FitType.eRecipCubeRoot:
                    return "1/" + sVar + "^1/3";
                    break;
                case TypeModule.FitType.eRecipSquare:
                    return "1/" + sVar + "^2";
                    break;
                case TypeModule.FitType.eRecipCube:
                    return "1/" + sVar + "^3";
                    break;
                case TypeModule.FitType.eSqrt:
                    return sVar + "^1/2";
                    break;
                case TypeModule.FitType.eRecipSqrt:
                    return "1/" + sVar + "^1/2";
                    break;
                case TypeModule.FitType.eLn:
                    return "Ln(" + sVar + ")";
                    break;
                default:
                    return "";
            }
        }

        private bool CanHandleZero(TypeModule.FitType eVal)
        {
            bool returnValue;

            switch (eVal)
            {
                case TypeModule.FitType.eLinear:
                    returnValue = true;
                    break;

                case TypeModule.FitType.eSquare:
                    returnValue = true;
                    break;

                case TypeModule.FitType.eCube:
                    returnValue = true;
                    break;

                case TypeModule.FitType.eCubeRoot:
                    returnValue = true;
                    break;

                case TypeModule.FitType.eSqrt:

                    returnValue = true;
                    break;
                default:

                    returnValue = false;
                    break;
            }
            return returnValue;
        }

        private bool CanHandleZeroAndNegative(TypeModule.FitType eVal)
        {
            bool returnValue;

            switch (eVal)
            {
                case TypeModule.FitType.eLinear:
                    returnValue = true;
                    break;

                case TypeModule.FitType.eSquare:
                    returnValue = true;
                    break;

                case TypeModule.FitType.eCube:

                    returnValue = true;
                    break;
                default:

                    returnValue = false;
                    break;
            }
            return returnValue;
        }

        private bool CanHandleNegative(TypeModule.FitType eVal)
        {
            bool returnValue;

            switch (eVal)
            {
                case TypeModule.FitType.eLinear:
                    returnValue = true;
                    break;

                case TypeModule.FitType.eSquare:
                    returnValue = true;
                    break;

                case TypeModule.FitType.eCube:
                    returnValue = true;
                    break;

                case TypeModule.FitType.eRecip:
                    returnValue = true;
                    break;

                case TypeModule.FitType.eRecipSquare:
                    returnValue = true;
                    break;

                case TypeModule.FitType.eRecipCube:

                    returnValue = true;
                    break;
                default:

                    returnValue = false;
                    break;
            }
            return returnValue;
        }


        private void CalcLR(ref CResults objRes, ref CErrors objErrs)
        {
            double A;
            double B;

            if (m_fSum < 2)
            {
                return;
            }
            // caluclate regression
            try
            {
                m_fMeanX1 = m_fSumX11 / m_fSum;
                m_fMeanX2 = m_fSumX21 / m_fSum;
                m_fMeanY = m_fSumY / m_fSum;
                m_fSdevX1 = Math.Sqrt((m_fSumX12 - Math.Pow(m_fSumX11, 2) / m_fSum) / (m_fSum - 1));
                m_fSdevX2 = Math.Sqrt((m_fSumX22 - Math.Pow(m_fSumX21, 2) / m_fSum) / (m_fSum - 1));
                m_fSdevY = Math.Sqrt((m_fSumY2 - Math.Pow(m_fSumY, 2) / m_fSum) / (m_fSum - 1));

                A = (m_fSum * m_fSumX12 - Math.Pow(m_fSumX11, 2)) * (m_fSum * m_fSumX2Y - m_fSumX21 * m_fSumY);
                B = (m_fSum * m_fSumX1X2 - m_fSumX11 * m_fSumX21) * (m_fSum * m_fSumX1Y - m_fSumX11 * m_fSumY);

                m_fSlope2 = (A - B) / (
                 (m_fSum * m_fSumX12 - m_fSumX11 * m_fSumX11) *
                 (m_fSum * m_fSumX22 - m_fSumX21 * m_fSumX21) -
                 Math.Pow(m_fSum * m_fSumX1X2 - m_fSumX11 * m_fSumX21, 2)
                 );
                m_fSlope1 = ((m_fSum * m_fSumX1Y - m_fSumX11 * m_fSumY) -
                 m_fSlope2 * (m_fSum * m_fSumX1X2 - m_fSumX11 * m_fSumX21)) /
                 (m_fSum * m_fSumX12 - m_fSumX11 * m_fSumX11);
                m_fIntercept = m_fMeanY - m_fSlope1 * m_fMeanX1 - m_fSlope2 * m_fMeanX2;
                m_fR2 = (m_fIntercept * m_fSumY + m_fSlope1 * m_fSumX1Y + m_fSlope2 * m_fSumX2Y -
                    m_fSumY * m_fSumY / m_fSum) / (m_fSumY2 - m_fSumY * m_fSumY / m_fSum);

            }
            catch (Exception ex)
            {
                objErrs.Add("Error in model " + m_sTY + "= A0 + A1 " + m_sTX1 + " + A2 " + m_sTX2 + "\r\n" + ex.Message.ToString());
            }
        }
    }

}
